<?php

namespace yifang\db\concern;

use Exception;
use yifang\utils\Str;
use yifang\db\Collection;

/**
 * 参数绑定支持
 */
trait Result
{
    /**
     * 设置数据处理（支持模型）
     */
    public function filter(callable $filter, $index = null)
    {
        if ($index) {
            $this->options['filter'][$index] = $filter;
        } else {
            $this->options['filter'][] = $filter;
        }
        return $this;
    }

    /**
     * 是否允许返回空数据（或空模型）
     */
    public function allowEmpty($allowEmpty = true)
    {
        $this->options['allow_empty'] = $allowEmpty;
        return $this;
    }

    /**
     * 设置查询数据不存在是否抛出异常
     */
    public function failException($fail = true)
    {
        $this->options['fail'] = $fail;
        return $this;
    }

    /**
     * 处理数据
     * @access protected
     * @param array $result 查询数据
     * @return void
     */
    protected function result(array &$result)
    {
        // JSON数据处理
        if (!empty($this->options['json'])) {
            $this->jsonResult($result);
        }
        // 查询数据处理
        foreach ($this->options['filter'] as $filter) {
            $result = call_user_func_array($filter, [$result, $this->options]);
        }
        // 获取器
        if (!empty($this->options['with_attr'])) {
            $this->getResultAttr($result, $this->options['with_attr']);
        }
    }

    /**
     * 查询失败 抛出异常
     */
    protected function throwNotFound()
    {
        if (!empty($this->model)) {
            $class = get_class($this->model);
            throw new Exception('model data Not Found:' . $class, $class, $this->options);
        }
        $table = $this->getTable();
        throw new Exception('table data not Found:' . $table, $table, $this->options);
    }

    /**
     * 处理空数据
     */
    protected function resultToEmpty()
    {
        if (!empty($this->options['fail'])) {
            $this->throwNotFound();
        } elseif (!empty($this->options['allow_empty'])) {
            return !empty($this->model) ? $this->model->newInstance() : [];
        }
    }

    /**
     * 处理数据集
     */
    protected function resultSet(array &$resultSet,$toCollection = true)
    {
        foreach ($resultSet as &$result) {
            $this->result($result);
        }
        // 返回Collection对象
        if ($toCollection) {
            $resultSet = new Collection($resultSet);
        }
    }

    /**
     * 使用获取器处理数据
     */
    protected function getResultAttr(array &$result, array $withAttr = [])
    {
        foreach ($withAttr as $name => $closure) {
            $name = Str::snake($name);
            if (strpos($name, '.')) {
                // 支持JSON字段 获取器定义
                list($key, $field) = explode('.', $name);

                if (isset($result[$key])) {
                    $result[$key][$field] = $closure($result[$key][$field] ?? null, $result[$key]);
                }
            } else {
                $result[$name] = $closure($result[$name] ?? null, $result);
            }
        }
    }

    /**
     * JSON字段数据转换
     * @access protected
     * @param array $result 查询数据
     * @return void
     */
    protected function jsonResult(array &$result)
    {
        foreach ($this->options['json'] as $name) {
            if (!isset($result[$name])) {
                continue;
            }
            $result[$name] = json_decode($result[$name], true);
        }
    }

    /**
     * 查找单条记录 不存在返回空数据（或者空模型）
     */
    public function findOrEmpty($data = null)
    {
        return $this->allowEmpty(true)->find($data);
    }
    /**
     * 查找多条记录 如果不存在则抛出异常
     */
    public function selectOrFail($data = null)
    {
        return $this->failException(true)->select($data);
    }

    /**
     * 查找单条记录 如果不存在则抛出异常
     */
    public function findOrFail($data = null)
    {
        return $this->failException(true)->find($data);
    }
}