<?php

namespace yifang\model\relation;

use yifang\model\Relation;
use Closure;
use yifang\utils\Str;
use yifang\db\BaseQuery as Query;
use yifang\model\Model;
use yifang\model\Pivot;
use yifang\db\Raw;
use Exception;

/**
 * 远程一对多关联类
 */
class HasManyThrough extends Relation
{
    /**
     * 中间关联表外键
     * @var string
     */
    protected $throughKey;

    /**
     * 中间主键
     * @var string
     */
    protected $throughPk;

    /**
     * 中间表查询对象
     * @var Query
     */
    protected $through;

    /**
     * 架构函数
     * @access public
     */
    public function __construct(Model $parent, $model,  $through,  $foreignKey,  $throughKey,  $localKey,  $throughPk)
    {
        $this->parent     = $parent;
        $this->model      = $model;
        $this->through    = (new $through)->db();
        $this->foreignKey = $foreignKey;
        $this->throughKey = $throughKey;
        $this->localKey   = $localKey;
        $this->throughPk  = $throughPk;
        $this->query      = (new $model)->db();
    }

    /**
     * 延迟获取关联数据
     * @return Collection
     */
    public function getRelation(array $subRelation = [], Closure $closure = null)
    {
        if ($closure) {
            $closure($this->getClosureType($closure));
        }
        $this->baseQuery();
        if ($this->withLimit) {
            $this->query->limit($this->withLimit);
        }
        return $this->query->relation($subRelation)->select()->setParent(clone $this->parent);
    }

    /**
     * 根据关联条件查询当前模型
     * @return Query
     */
    public function has($operator = '>=', $count = 1,  $id = '*',  $joinType = '', Query $query = null)
    {
        $model         = Str::snake(class_basename($this->parent));
        $throughTable  = $this->through->getTable();
        $pk            = $this->throughPk;
        $throughKey    = $this->throughKey;
        $relation      = new $this->model;
        $relationTable = $relation->getTable();
        $softDelete    = $this->query->getOptions('soft_delete');

        if ('*' != $id) {
            $id = $relationTable . '.' . $relation->getPk();
        }
        $query = $query ?: $this->parent->db()->alias($model);

        return $query->field($model . '.*')
            ->join($throughTable, $throughTable . '.' . $this->foreignKey . '=' . $model . '.' . $this->localKey)
            ->join($relationTable, $relationTable . '.' . $throughKey . '=' . $throughTable . '.' . $this->throughPk)
            ->when($softDelete, function ($query) use ($softDelete, $relationTable) {
                $query->where($relationTable . strstr($softDelete[0], '.'), '=' == $softDelete[1][0] ? $softDelete[1][1] : null);
            })
            ->group($relationTable . '.' . $this->throughKey)
            ->having('count(' . $id . ')' . $operator . $count);
    }

    /**
     * 根据关联条件查询当前模型
     * @return Query
     */
    public function hasWhere($where = [], $fields = null, $joinType = '', Query $query = null)
    {
        $model        = Str::snake(class_basename($this->parent));
        $throughTable = $this->through->getTable();
        $pk           = $this->throughPk;
        $throughKey   = $this->throughKey;
        $modelTable   = (new $this->model)->getTable();

        if (is_array($where)) {
            $this->getQueryWhere($where, $modelTable);
        } elseif ($where instanceof Query) {
            $where->via($modelTable);
        } elseif ($where instanceof Closure) {
            $where($this->query->via($modelTable));
            $where = $this->query;
        }

        $fields     = $this->getRelationQueryFields($fields, $model);
        $softDelete = $this->query->getOptions('soft_delete');
        $query      = $query ?: $this->parent->db();

        return $query->alias($model)
            ->join($throughTable, $throughTable . '.' . $this->foreignKey . '=' . $model . '.' . $this->localKey)
            ->join($modelTable, $modelTable . '.' . $throughKey . '=' . $throughTable . '.' . $this->throughPk, $joinType)
            ->when($softDelete, function ($query) use ($softDelete, $modelTable) {
                $query->where($modelTable . strstr($softDelete[0], '.'), '=' == $softDelete[1][0] ? $softDelete[1][1] : null);
            })
            ->group($modelTable . '.' . $this->throughKey)
            ->where($where)
            ->field($fields);
    }

    /**
     * 预载入关联查询（数据集）
     * @return void
     */
    public function eagerlyResultSet(array &$resultSet,  $relation, array $subRelation = [], Closure $closure = null, array $cache = [])
    {
        $localKey   = $this->localKey;
        $foreignKey = $this->foreignKey;
        $range = [];
        foreach ($resultSet as $result) {
            // 获取关联外键列表
            if (isset($result->$localKey)) {
                $range[] = $result->$localKey;
            }
        }
        if (!empty($range)) {
            $this->query->removeWhereField($foreignKey);

            $data = $this->eagerlyWhere([
                [$this->foreignKey, 'in', $range],
            ], $foreignKey, $subRelation, $closure, $cache);
            // 关联数据封装
            foreach ($resultSet as $result) {
                $pk = $result->$localKey;
                if (!isset($data[$pk])) {
                    $data[$pk] = [];
                }
                // 设置关联属性
                $result->setRelation($relation, $this->resultSetBuild($data[$pk], clone $this->parent));
            }
        }
    }

    /**
     * 预载入关联查询（数据）
     * @return void
     */
    public function eagerlyResult(Model $result,  $relation, array $subRelation = [], Closure $closure = null, array $cache = [])
    {
        $localKey   = $this->localKey;
        $foreignKey = $this->foreignKey;
        $pk         = $result->$localKey;
        $this->query->removeWhereField($foreignKey);
        $data = $this->eagerlyWhere([
            [$foreignKey, '=', $pk],
        ], $foreignKey, $subRelation, $closure, $cache);
        // 关联数据封装
        if (!isset($data[$pk])) {
            $data[$pk] = [];
        }
        $result->setRelation($relation, $this->resultSetBuild($data[$pk], clone $this->parent));
    }

    /**
     * 关联模型预查询
     * @return array
     */
    protected function eagerlyWhere(array $where,  $key, array $subRelation = [], Closure $closure = null, array $cache = [])
    {
        // 预载入关联查询 支持嵌套预载入
        $throughList = $this->through->where($where)->select();
        $keys        = $throughList->column($this->throughPk, $this->throughPk);
        if ($closure) {
            $this->baseQuery = true;
            $closure($this->getClosureType($closure));
        }
        $throughKey = $this->throughKey;
        if ($this->baseQuery) {
            $throughKey = Str::snake(class_basename($this->model)) . "." . $this->throughKey;
        }
        $list = $this->query
            ->where($throughKey, 'in', $keys)
            ->cache($cache[0] ?? false, $cache[1] ?? null, $cache[2] ?? null)
            ->select();

        // 组装模型数据
        $data = [];
        $keys = $throughList->column($this->foreignKey, $this->throughPk);
        foreach ($list as $set) {
            $key = $keys[$set->{$this->throughKey}];

            if ($this->withLimit && isset($data[$key]) && count($data[$key]) >= $this->withLimit) {
                continue;
            }
            $data[$key][] = $set;
        }
        return $data;
    }

    /**
     * 关联统计
     * @return mixed
     */
    public function relationCount(Model $result, Closure $closure = null,  $aggregate = 'count',  $field = '*',  &$name = null)
    {
        $localKey = $this->localKey;
        if (!isset($result->$localKey)) {
            return 0;
        }
        if ($closure) {
            $closure($this->getClosureType($closure), $name);
        }
        $alias        = Str::snake(class_basename($this->model));
        $throughTable = $this->through->getTable();
        $pk           = $this->throughPk;
        $throughKey   = $this->throughKey;
        $modelTable   = $this->parent->getTable();
        if (false === strpos($field, '.')) {
            $field = $alias . '.' . $field;
        }
        return $this->query
            ->alias($alias)
            ->join($throughTable, $throughTable . '.' . $pk . '=' . $alias . '.' . $throughKey)
            ->join($modelTable, $modelTable . '.' . $this->localKey . '=' . $throughTable . '.' . $this->foreignKey)
            ->where($throughTable . '.' . $this->foreignKey, $result->$localKey)
            ->$aggregate($field);
    }

    /**
     * 创建关联统计子查询
     * @return string
     */
    public function getRelationCountQuery(Closure $closure = null,  $aggregate = 'count',  $field = '*',  &$name = null)
    {
        if ($closure) {
            $closure($this->getClosureType($closure), $name);
        }
        $alias        = Str::snake(class_basename($this->model));
        $throughTable = $this->through->getTable();
        $pk           = $this->throughPk;
        $throughKey   = $this->throughKey;
        $modelTable   = $this->parent->getTable();
        if (false === strpos($field, '.')) {
            $field = $alias . '.' . $field;
        }
        return $this->query
            ->alias($alias)
            ->join($throughTable, $throughTable . '.' . $pk . '=' . $alias . '.' . $throughKey)
            ->join($modelTable, $modelTable . '.' . $this->localKey . '=' . $throughTable . '.' . $this->foreignKey)
            ->whereExp($throughTable . '.' . $this->foreignKey, '=' . $this->parent->getTable() . '.' . $this->localKey)
            ->fetchSql()
            ->$aggregate($field);
    }

    /**
     * 执行基础查询（仅执行一次）
     * @access protected
     * @return void
     */
    protected function baseQuery()
    {
        if (empty($this->baseQuery) && $this->parent->getData()) {
            $alias        = Str::snake(class_basename($this->model));
            $throughTable = $this->through->getTable();
            $pk           = $this->throughPk;
            $throughKey   = $this->throughKey;
            $modelTable   = $this->parent->getTable();
            if ($this->withoutField) {
                $this->query->withoutField($this->withoutField);
            }
            $fields = $this->getQueryFields($alias);
            $this->query
                ->field($fields)
                ->alias($alias)
                ->join($throughTable, $throughTable . '.' . $pk . '=' . $alias . '.' . $throughKey)
                ->join($modelTable, $modelTable . '.' . $this->localKey . '=' . $throughTable . '.' . $this->foreignKey)
                ->where($throughTable . '.' . $this->foreignKey, $this->parent->{$this->localKey});

            $this->baseQuery = true;
        }
    }

}
